<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:math="http://www.w3.org/2005/xpath-functions/math" xmlns:map="http://www.w3.org/2005/xpath-functions/map"
    xmlns:array="http://www.w3.org/2005/xpath-functions/array" xmlns:doc="http://www.saxonica.com/ns/doc" xmlns:f="MyFunctions"
    xmlns:svg="http://www.w3.org/2000/svg" xmlns:saxon="http://saxon.sf.net/" xmlns:file="http://expath.org/ns/file" exclude-result-prefixes="xs math"
    version="3.0">



    <doc:synopsis title="Statistics report generator">
        <p>This stylesheet is usually invoked to postprocess a statistics generation run for rule pattern optimization. It can be invoked separately, but you
            shouldn't need to in normal circumstances.</p>
        <p>
            <b>The results from this stylesheet employ resources that are set up by <code>com.saxonica.StatsTransform.java</code></b>
        </p>
    </doc:synopsis>

    <xsl:import href="stats.barchart.xsl"/>
    <xsl:import href="stats-rank.xsl"/>
    <xsl:import href="stats-time.xsl"/>
    <xsl:import href="stats-call.xsl"/>
    <xsl:import href="stats-html.xsl"/>
    <xsl:include href="stats.svg.xsl"/>
    <xsl:include href="map-print.xsl"/>
    <xsl:include href="stats.table2text.xsl"/>

    <xsl:param name="xsl" doc:doc="The default stylesheet that was used during the runs">Undefined</xsl:param>



    <xsl:include href="stats.process.xsl"/>

    <!-- *IMPORTANT* This needs the '-l' option on Saxon to track source line numbers -->

    <!-- This process will produce a subdirectory and an HTML that will point to resources in that subdirectory -->

    <doc:synopsis title="Stats log sets preprocessing">
        <p>Preprocess a set of statistics log files</p>
    </doc:synopsis>
    <xsl:output method="xml" indent="yes" saxon:indent-spaces="2"/>
    <xsl:mode on-no-match="shallow-copy"/>

    <xsl:variable name="this.base-uri" select="base-uri(doc(''))"/>
    <xsl:variable name="base-uri" select="base-uri()"/>

    <xsl:variable name="base-name" select="replace(tokenize($base-uri, '/')[last()], '.xml$', '')"/>
    <xsl:variable name="dir">statistics</xsl:variable>

    <xsl:variable name="dir2" select="resolve-uri($dir, $base-uri)"/>

    <xsl:variable name="lib" select="resolve-uri('../', $this.base-uri)"/>
    <xsl:variable name="tools.lib" select="resolve-uri('../lib/', $this.base-uri)"/>

    <xsl:variable name="graphics.css">graphics.css</xsl:variable>
    <xsl:variable name="new.graphics.css" select="
            string-join(($dir2,
            $graphics.css), '/')"/>
    <xsl:variable name="out" select="
            string-join(($dir,
            $base-name), '/')"/>

    <xsl:variable name="built-in.mode-defaults.base" as="map(*)" doc:doc="Base built-in rule sets"
        select="map {
        'DC' : 'built-in: deep-copy',
        'DS' : 'built-in: deep-skip',
        'FF' : 'built-in: fail',
        'SC' : 'built-in: shallow-copy',
        'SS' : 'built-in: shallow-skip',
        'TC' : 'built-in: text-only-copy'}
        "/>
    <xsl:variable name="built-in.mode-defaults" as="map(*)" doc:doc="Built-in rule set labels, including warning derivatives"
        select="
            map:merge(($built-in.mode-defaults.base,
            map:for-each($built-in.mode-defaults.base, function ($k, $v) {
                map:entry($k || '+W', $v)
            })))"/>

    <xsl:variable name="colours" as="xs:string*" doc:doc="Colours used for graph styling"
        select="
            'red',
            'blue',
            'green',
            'lightblue',
            'orange',
            'cyan',
            'brown'"/>

    <xsl:template match="/">
        <xsl:message select="'xsl=', $xsl"/>
        <xsl:variable name="stats.docs"
            select="
                //case/doc(resolve-uri(
                concat(@dir, (@stats,
                'stats.xml')[1]), base-uri(.)))/*"/>
        <xsl:variable name="p1">
            <stats uri="{base-uri()}">
                <xsl:apply-templates select="stats/(@xsl | @s | @o | @repeat | @description)"/>
                <xsl:sequence select="$stats.docs/(@xsl | @s | @o | @repeat)"/>
                <xsl:apply-templates select="stats/group">
                    <xsl:with-param name="default-xsl" select="$xsl" tunnel="yes"/>
                </xsl:apply-templates>
            </stats>
        </xsl:variable>
        <xsl:variable name="big-modes" select="$p1/stats/group/comparison/mode[@isBig = true()]/@name"/>
        <xsl:variable name="case.names" select="distinct-values($p1//case/@name)" as="xs:string*"/>

        <xsl:variable name="css.parts" as="xs:string" doc:doc="produce CSS entries mapping colours to the cases"
            select="
                string-join(for $i in 1 to count($case.names)
                return
                    '.' || $case.names[$i] || '{&#xA; stroke:' || $colours[$i] || ';&#xA; fill:' || $colours[$i] || ';&#xA;}', '&#xA;')"/>
        <!--  <xsl:message>Preparing styletools</xsl:message>-->
        <xsl:call-template name="prepare">
            <xsl:with-param name="extra.css" select="$css.parts"/>
        </xsl:call-template>
        <!--  <xsl:message>Styletools prepared</xsl:message>-->
        <xsl:variable name="CSS" select="
                string-join((resolve-uri($dir, $base-uri),
                'baseStyle.css'), '/')"/>

        <xsl:variable name="modes" select="
                map:merge($p1/stats/group/stylesheet/mode/map:entry(@name, template))"/>
        <xsl:variable name="sources"
            select="
                distinct-values(map:for-each($modes, function ($k, $v) {
                    $v/@module
                }))"
            as="xs:string*"/>
        <xsl:variable name="sources.templates" select="map:merge($sources ! map:entry(., map:merge(doc(.)//xsl:template/map:entry(saxon:line-number(.), .))))"/>
        <xsl:variable name="sources.matches"
            select="
                map:merge(map:for-each($sources.templates,
                function ($k, $v) {
                    map:entry($k, map:merge(map:for-each($v, function ($k1, $v1) {
                        map:entry($k1, $v1/@match)
                    })))
                }))"/>

        <!--<xsl:for-each select="map:keys($sources.matches)">
            <xsl:message
                select=".,map:for-each($sources.matches(.),function($k,$v) {$k || ':' ||$v})"/>
        </xsl:for-each>-->

        <result>
            <xsl:sequence select="$p1"/>
            <xsl:call-template name="map:print">
                <xsl:with-param name="items" select="$sources.matches"/>
            </xsl:call-template>
            <a/>
            <xsl:variable name="ranks" as="element()*">
                <xsl:apply-templates select="$p1" mode="stats.rank.svg">
                    <xsl:with-param name="svg.root" select="$out" tunnel="yes"/>
                    <xsl:with-param name="modes.wanted" select="$big-modes" tunnel="yes"/>
                    <xsl:with-param name="sources.matches" as="map(*)" select="$sources.matches" tunnel="yes"/>
                    <xsl:with-param name="sources.templates" as="map(*)" select="$sources.templates" tunnel="yes"/>
                    <xsl:with-param name="CSS.file" select="$new.graphics.css" tunnel="yes"/>
                </xsl:apply-templates>
            </xsl:variable>
            <xsl:variable name="time" as="element()*">
                <xsl:apply-templates select="$p1" mode="stats.time.svg">
                    <xsl:with-param name="svg.root" select="$out" tunnel="yes"/>
                    <xsl:with-param name="sources.matches" as="map(*)" select="$sources.matches" tunnel="yes"/>
                    <xsl:with-param name="sources.templates" as="map(*)" select="$sources.templates" tunnel="yes"/>
                    <xsl:with-param name="CSS.file" select="$new.graphics.css" tunnel="yes"/>
                </xsl:apply-templates>
            </xsl:variable>
            <xsl:variable name="call" as="element()*">
                <xsl:apply-templates select="$p1" mode="stats.call.svg">
                    <xsl:with-param name="svg.root" select="$out" tunnel="yes"/>
                    <xsl:with-param name="sources.matches" as="map(*)" select="$sources.matches" tunnel="yes"/>
                    <xsl:with-param name="sources.templates" as="map(*)" select="$sources.templates" tunnel="yes"/>
                    <xsl:with-param name="CSS.file" select="$new.graphics.css" tunnel="yes"/>
                </xsl:apply-templates>
            </xsl:variable>

            <xsl:variable name="time.table" as="element()*">
                <xsl:apply-templates select="$p1" mode="stats.time.table">
                    <xsl:with-param name="sources.matches" as="map(*)" select="$sources.matches" tunnel="yes"/>
                    <xsl:with-param name="sources.templates" as="map(*)" select="$sources.templates" tunnel="yes"/>
                </xsl:apply-templates>
            </xsl:variable>

            <xsl:apply-templates select="$p1" mode="stats.html" use-when="true()">
                <xsl:with-param name="svg.root" select="substring-after($out, '/')" tunnel="yes"/>
                <xsl:with-param name="name" select="$out"/>
                <xsl:with-param name="svg" select="$time, $call, $ranks" use-when="true()"/>
                <xsl:with-param name="table" select="$time.table"/>
            </xsl:apply-templates>
            <xsl:for-each select="$time, $call, $ranks" use-when="true()">
                <xsl:call-template name="svg:makeSVGFile">
                    <xsl:with-param name="svg.root" select="$out"/>
                    <xsl:with-param name="mode" select="substring-before(@id, '/')"/>
                    <xsl:with-param name="name" select="substring-after(@id, '/')"/>
                    <xsl:with-param name="svg" select="."/>
                </xsl:call-template>
            </xsl:for-each>

            <xsl:result-document href="tempCSS.xsl" use-when="false()">
                <xsl:variable name="stylesheet.attributes" as="attribute()*" expand-text="no">
                    <xsl:attribute name="xpath-default-namespace">http://www.w3.org/2000/svg</xsl:attribute>
                </xsl:variable>
                <xsl:call-template name="CSS2XSLT">
                    <xsl:with-param name="URI" select="xs:anyURI($new.graphics.css)"/>
                    <xsl:with-param name="stylesheet.attributes" select="$stylesheet.attributes"/>
                </xsl:call-template>
            </xsl:result-document>

        </result>
        <xsl:message>**** Data analysis completed</xsl:message>

        <xsl:variable name="case-stats" as="element()*">
            <xsl:apply-templates select="$p1//group/case" mode="stats.html"/>
        </xsl:variable>

        <xsl:message>&#xA;** Overall comparison figures:</xsl:message>
        <xsl:for-each select="f:table2text(($case.table.header/*, $case-stats))">
            <xsl:message select="'   ', ."/>
        </xsl:for-each>

        <xsl:message>&#xA;**** Result webpage constructed:</xsl:message>
        <xsl:message>**** <xsl:value-of select="resolve-uri($out || '.html', base-uri())"/></xsl:message>

    </xsl:template>



    <xsl:template name="prepare" doc:doc="Copy the necessary library files, including SaxonCE, into the target destination.">
        <xsl:param name="extra.css" as="xs:string" doc:doc="Additional entries to add to the graphics CSS file"/>
        <!-- Much of this is now handled in the calling StatsTransform.java-->
        <xsl:variable name="newCSS" select="string-join(($dir2, $graphics.css), '/')"/>
        <xsl:sequence
            select="
                if (not(file:exists($dir2))) then
                    file:create-dir($dir2)
                else
                    ()"/>
        <!--<xsl:sequence select="('Saxonce') ! file:copy(resolve-uri(., $lib), $dir2)"/>
        <xsl:sequence select="('style',$graphics.css) ! file:copy(resolve-uri(., $this.base-uri), $dir2)"/>-->
        <xsl:sequence select="file:append-text($newCSS, '&#xA;' || $extra.css)"/>
    </xsl:template>

    <xsl:template match="@xsl | @s | @o">
        <xsl:attribute name="{name()}" select="resolve-uri(., base-uri())"/>
    </xsl:template>
    <xsl:template match="@repeat">
        <xsl:sequence select="."/>
    </xsl:template>

    <xsl:template match="group"
        doc:doc="Collect corresponding samples for each rule in a comparison element,
        subdivided by mode, placing a case element for each sample under a rule.
       Modes are judged as to whether they are considered 'big'.">
        <xsl:copy>
            <xsl:attribute name="svg" select="$base-name"/>
            <xsl:apply-templates select="@*"/>
            <xsl:variable name="stylesheets" as="element()*">
                <xsl:apply-templates select="case" mode="show.stylesheets"/>
            </xsl:variable>
            <xsl:sequence
                select="
                    for $x in distinct-values($stylesheets/@xsl)
                    return
                        $stylesheets[@xsl = $x][1]"/>
            <xsl:variable name="cases" as="element()*">
                <xsl:apply-templates select="case"/>
            </xsl:variable>
            <xsl:for-each select="$cases">
                <xsl:copy>
                    <xsl:sequence select="@*"/>
                    <xsl:attribute name="total.pattern" select="format-number(stats/@total.time, '0')"/>
                    <xsl:sequence select="* except stats"/>
                </xsl:copy>
            </xsl:for-each>
            <xsl:variable name="times" select="$cases/stats/mode/rule"/>
            <comparison>
                <xsl:for-each-group select="$times" group-by="../@name">
                    <mode name="{current-grouping-key()}">
                        <xsl:variable name="temp" as="element()*">
                            <xsl:for-each-group select="current-group()" group-by="@loc">
                                <!--<xsl:variable name="line" select="number(substring-after(current-grouping-key(), '#'))"/>
                                <xsl:variable name="module" select="substring-before(current-grouping-key(), '#')"/>-->
                                <rule>
                                    <xsl:sequence select="@built-in, @loc, @seq, @rank, @prec"/>
                                    <xsl:attribute name="max" select="xs:integer(max(current-group()/@times.total))"/>
                                    <xsl:attribute name="invocations" select="xs:integer(max(current-group()/@invocations))"/>
                                    <xsl:for-each select="current-group()">
                                        <!--<xsl:sort select="../../../@name"/>-->
                                        <case>
                                            <xsl:sequence select="@* except (@loc, @mode, @seq, @rank), ../../../@name"/>
                                        </case>
                                    </xsl:for-each>
                                </rule>
                            </xsl:for-each-group>
                        </xsl:variable>
                        <xsl:variable name="invocations" select="xs:integer(sum($temp/@invocations))"/>
                        <xsl:attribute name="max.time" select="xs:integer(max($temp/@max))"/>
                        <xsl:attribute name="invocations" select="$invocations"/>
                        <xsl:attribute name="isBig" select="$invocations gt 1000"/>
                        <xsl:for-each select="$temp">
                            <xsl:sort select="number(@max)" order="descending"/>
                            <xsl:copy-of select="."/>
                        </xsl:for-each>
                    </mode>
                </xsl:for-each-group>
            </comparison>
        </xsl:copy>
    </xsl:template>

    <xsl:template match="case" as="element()*" mode="show.stylesheets">
        <xsl:param name="default-xsl" tunnel="yes"/>
        <xsl:variable name="dir" select="@dir"/>
        <xsl:variable name="explain" select="
                resolve-uri(concat($dir, (@explain, 'explain.xml')[1]), base-uri(.))"/>
        <xsl:variable name="explain.docs" select="doc($explain)"/>
        <xsl:variable name="sources" select="distinct-values($explain.docs//templateRule/@module)" as="xs:string*"/>
        <xsl:variable name="stylesheet.templates" select="map:merge($sources ! map:entry(., doc(.)//xsl:template[@match]))"/>
        <stylesheet xsl="{(@xsl,$default-xsl)[1]}">
            <xsl:for-each select="$explain.docs/stylesheet/mode">
                <mode name="{(@name,'xsl:unnamed')[1]}" n.rules="{count(.//templateRule)}">
                    <xsl:apply-templates select=".//templateRule">
                        <xsl:with-param name="stylesheet.templates" select="$stylesheet.templates" tunnel="yes"/>
                        <xsl:sort select="number(@rank)" order="descending"/>
                        <xsl:sort select="number(@seq)" order="descending"/>
                    </xsl:apply-templates>
                </mode>
            </xsl:for-each>
        </stylesheet>
    </xsl:template>

    <xsl:template match="templateRule">
        <xsl:param name="stylesheet.templates" as="map(*)" tunnel="yes"/>
        <xsl:variable name="attributes"
            select="
                @module, @line, @rank, @seq, @prec,
                ancestor::ruleSet/@type,
                parent::ruleChain/@key"/>
        <xsl:variable name="template" select="$stylesheet.templates(@module)[saxon:line-number(.) = current()/@line ]"/>
        <xsl:for-each select="$template">
            <template match="{@match}">
                <xsl:sequence select="$attributes"/>
            </template>
        </xsl:for-each>
    </xsl:template>

    <xsl:template match="case" as="element()">
        <xsl:variable name="dir" select="@dir"/>
        <xsl:variable name="stats" select="
                resolve-uri(
                concat($dir, (@stats,
                'stats.xml')[1]), base-uri(.))"/>
        <xsl:variable name="stats.docs" select="doc($stats)"/>

        <xsl:copy>
            <xsl:attribute name="xml:base" select="base-uri()"/>
            <xsl:sequence select="@*"/>
            <xsl:attribute name="stats" select="$stats"/>
            <xsl:attribute name="name" select="'case' || position()"/>
            <xsl:apply-templates select="@label" mode="config"/>
            <xsl:apply-templates select="$stats.docs" mode="config"/>
            <!-- <xsl:sequence select="
                    @name,
                    @label"/>-->
            <xsl:apply-templates select="$stats.docs" mode="process.stats"/>
            <xsl:variable name="explain"
                select="
                    resolve-uri(
                    concat($dir, (@explain,
                    'explain.xml')[1]), base-uri(.))"/>
            <xsl:variable name="explain.docs" select="doc($explain)"/>
            <xsl:sequence select="$explain.docs//elementRuleIndexes"/>

        </xsl:copy>
    </xsl:template>

    <xsl:mode name="config" on-no-match="deep-skip"/>
    <xsl:template match="document-node() | statsRTL | config" mode="config">
        <xsl:apply-templates select="@* | node()" mode="#current"/>
    </xsl:template>
    <xsl:template match="
            statsRTL/@total.time | statsRTL/@dateTime |
            config/patternOptEE | config/@edition | config/@version"
        mode="config">
        <xsl:copy-of select="."/>
    </xsl:template>
    <xsl:template match="case/@label | config/@label" mode="config">
        <xsl:sequence select="."/>
        <xsl:variable name="parts" select="tokenize(., '\s*;\s*')"/>
        <xsl:if test="count($parts) gt 2">
            <xsl:attribute name="name" select="$parts[1]"/>
            <xsl:attribute name="label" select="$parts[2]"/>
        </xsl:if>
    </xsl:template>

</xsl:stylesheet>
