<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:math="http://www.w3.org/2005/xpath-functions/math"
    xmlns:map="http://www.w3.org/2005/xpath-functions/map"
    xmlns:array="http://www.w3.org/2005/xpath-functions/array"
    xmlns:doc="http://www.saxonica.com/ns/doc" xmlns:f="MyFunctions"
    xmlns:svg="http://www.w3.org/2000/svg" xmlns:saxon="http://saxon.sf.net/"
    xmlns:file="http://expath.org/ns/file" exclude-result-prefixes="xs math" version="3.0">

    <doc:synopsis title="Table to Text">A library of routines that convert HTML tables into text
        lines</doc:synopsis>

    <xsl:function name="f:table2text" as="item()*"
        doc:doc="Convert a set of HTML table rows to a series of text strings, with appropriate spacing">
        <xsl:param name="rows" as="element()*"/>
        <xsl:variable name="n.columns" select="max($rows/count(*))"/>
        <xsl:variable name="new-rows" as="element()*">
            <xsl:iterate select="$rows">
                <xsl:param name="rowspans" as="element()*">
                    <xsl:for-each select="1 to $n.columns">
                        <td rowspan="1"/>
                    </xsl:for-each>
                </xsl:param>
                <xsl:variable name="new-cells" select="f:table.rowspan($rowspans, *)"/>
                <xsl:copy>
                    <xsl:sequence select="@*, $new-cells"/>
                </xsl:copy>
                <xsl:next-iteration>
                    <xsl:with-param name="rowspans" select="$new-cells"/>
                </xsl:next-iteration>
            </xsl:iterate>
        </xsl:variable>
        <xsl:variable name="column.widths"
            select="
                map:merge(
                for $i in 1 to $n.columns
                return
                    map:entry($i, max($new-rows/*[$i]/string-length())))"/>
        <!--xsl:value-of select=" (1 to $n.columns) ! $column.widths(.)" separator=","/>-->
        <xsl:for-each select="$new-rows">
            <xsl:value-of>
                <xsl:for-each select="*">
                    <xsl:value-of
                        select=".,string-join((1 to $column.widths(position()) - string-length())!' ','')"
                        separator="  "/>
                </xsl:for-each>
            </xsl:value-of>
        </xsl:for-each>
    </xsl:function>

    <xsl:function name="f:table.rowspan" as="element()*"
        doc:doc="Merge rowspans, by adding dummy cells">
        <xsl:param name="columns" as="element()*"/>
        <xsl:param name="cells" as="element()*"/>
        <xsl:choose>
            <xsl:when test="exists($columns)">
                <xsl:variable name="column" select="head($columns)"/>
                <xsl:choose>
                    <xsl:when test="$column/@rowspan and number($column/@rowspan) gt 1">
                        <td rowspan="{$column/@rowspan - 1}"/>
                        <xsl:sequence select="f:table.rowspan(tail($columns), $cells)"/>
                    </xsl:when>
                    <xsl:otherwise>
                        <xsl:sequence
                            select="head($cells), f:table.rowspan(tail($columns), tail($cells))"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:when>
            <xsl:when test="exists($cells)">
                <xsl:sequence select="$cells"/>
            </xsl:when>
        </xsl:choose>
    </xsl:function>
</xsl:stylesheet>
